  '=============================================================================
  '=                                TopDown 3D                                 =
  '=                                                                           =
  '=                        FX and particles functions                         =
  '=============================================================================

  type tExplosion
    render as byte
    x as single
    y as single
    z as single
    bSize as single   
    mSize as single ' Max size
  end type
  %LIGHTSOURCE_EXPLOSION   = %GL_LIGHT1
  %MAX_EXPLOSION_PARTICLES = 64*4
  global ExploPoly(%MAX_EXPLOSION_PARTICLES) as tExplosion  
  global ExploPolyID as long = 1
  global LightSyncBillboard as long

  type tBillboard
    render as byte  
    t     as byte     ' texture
    x     as single
    y     as single
    z     as single
    bSize as single  
    bSizeMax as single
    bSizeMin as single
    startTime as long ' ms
    endTime as long ' ms
  end type

  %MAX_BILLBOARDS = 257
  global Billboard(%MAX_BILLBOARDS) as tBillboard  
  global BillboardID as long
  
  ' -- Few useful functions form OpenGL headers
  %GL_MODELVIEW_MATRIX                 = &H0BA6
  DECLARE SUB glGetFloatv LIB "opengl32.dll" ALIAS "glGetFloatv" (BYVAL pname AS dword, BYREF params() AS single)
  DECLARE SUB glLoadMatrixf LIB "opengl32.dll" ALIAS "glLoadMatrixf" (BYREF m() AS single)   


  sub Effects_Init()
    tbgl_SetLightParameter %LIGHTSOURCE_EXPLOSION, %TBGL_LIGHT_DIFFUSE, 2, 0.5, 0, 1
    tbgl_SetLightParameter %LIGHTSOURCE_EXPLOSION, %TBGL_LIGHT_ambient, 0, 0, 0, 0
    tbgl_SetLightParameter %LIGHTSOURCE_EXPLOSION, %TBGL_LIGHT_position, 0, 0, -50, 1
  
    %TEX_EFFECT_FLAME     = 220
    %TEX_EFFECT_SHOT = 220
    %TEX_EFFECT_EXPLOSION = 221
    %TEX_EFFECT_SMOKE = 222
    %TEX_EFFECT_GLOW = 223
    
    rem tbgl_LoadTexture APP_SOURCEPATH + dirTextures + "flame_Stripe.bmp", %TEX_EFFECT_FLAME, %TBGL_TEX_MIPMAP
    tbgl_LoadTexture APP_SOURCEPATH + dirTextures + "fx_explosion.bmp" , %TEX_EFFECT_EXPLOSION, %TBGL_TEX_MIPMAP
    tbgl_LoadTexture APP_SOURCEPATH + dirTextures + "fx_smoke.bmp" , %TEX_EFFECT_SMOKE, %TBGL_TEX_MIPMAP
    tbgl_LoadTexture APP_SOURCEPATH + dirTextures + "fx_glow.bmp" , %TEX_EFFECT_GLOW, %TBGL_TEX_MIPMAP
    
    %LIST_EFFECT_FLAME     = 220
    %LIST_EFFECT_QUAD      = 221
    
    tbgl_NewList %LIST_EFFECT_FLAME

      tbgl_Pushmatrix
      tbgl_Scale 0.5,0.5,2
      tbgl_Rotate 90,-1,0,0

      tbgl_Sphere 1

      tbgl_Popmatrix  
      
    tbgl_EndList        

    tbgl_NewList %LIST_EFFECT_QUAD
    tbgl_BeginPoly %GL_QUADS
      tbgl_Normal 0,0,1
      tbgl_TexCoord2d 0,0 
      tbgl_Vertex -1,-1,0
      tbgl_TexCoord2d 1,0 
      tbgl_Vertex  1,-1,0
      tbgl_TexCoord2d 1,1 
      tbgl_Vertex  1, 1,0
      tbgl_TexCoord2d 0,1 
      tbgl_Vertex -1,1, 0
    tbgl_EndPoly  
    tbgl_EndList

  end sub
  
  sub Effects_Reset()
    local i as long
    
    for i = 1 to %MAX_EXPLOSION_PARTICLES      
      ExploPoly(i).render = 0 
    next  
    
    for i = 1 to %MAX_BILLBOARDS
      Billboard(i).render = 0 
    next    
  end sub
  
  ' Help function, which will generate new explosion for us
  sub Explosion_New( x as single, y as single, z as single, maxSize as single )
    local i as long
    tbgl_UseLightSource %LIGHTSOURCE_EXPLOSION, 1
    tbgl_SetLightParameter %LIGHTSOURCE_EXPLOSION, %TBGL_LIGHT_position, x, y, z, 1 
    for i = 1 to 12
      ExploPolyID = iif(ExploPolyID = %MAX_EXPLOSION_PARTICLES, 1, ExploPolyID+1)
      if i = 1 then LightSyncBillboard = ExploPolyID
      
      ExploPoly(ExploPolyID).x = x+cos(rndf(1,7)) * rndf(1,maxSize*2/3)
      ExploPoly(ExploPolyID).y = y+sin(rndf(1,7)) * rndf(1,maxSize*2/3)
      ExploPoly(ExploPolyID).z = z-sin(rndf(1,7)) * rndf(1,maxSize*2/3)
      ExploPoly(ExploPolyID).bSize = rndf(1,maxSize)         
      ExploPoly(ExploPolyID).mSize = maxSize         
      ExploPoly(ExploPolyID).render = 1          
    next
    
  end sub

  ' Explosion render
  sub Explosion_RenderAll ()
    local i as long
    local ColorDivisor as single

    tbgl_UseTexture 1
    tbgl_bindTexture %TEX_EFFECT_EXPLOSION

    tbgl_UseDepthmask 0
    tbgl_UseBlend 1
    tbgl_Uselighting 0
  
    for i = 1 to %MAX_EXPLOSION_PARTICLES
      
      if ExploPoly(i).render = 0 then iterate for
      
      ExploPoly(i).bSize += 15/FrameRate
      if ExploPoly(i).bSize >  ExploPoly(i).mSize then 
        ExploPoly(i).render = 0
       rem if i = LightSyncBillboard then tbgl_UseLightSource %LIGHTSOURCE_EXPLOSION, 0
        iterate for
      end if  

      tbgl_PushMatrix
        tbgl_Translate ExploPoly(i).x, ExploPoly(i).y, ExploPoly(i).z
        ColorDivisor = 256-ExploPoly(i).bSize*(256/ExploPoly(i).mSize)
        tbgl_color ColorDivisor, ColorDivisor, ColorDivisor
        if i = LightSyncBillboard then tbgl_SetLightParameter %LIGHTSOURCE_EXPLOSION, %TBGL_LIGHT_Diffuse, ColorDivisor/256, 0.5*ColorDivisor/256, 0, 1', ColorDivisor/512, 0, 0.5 
    
        billboard_Begin        
        tbgl_Scale ExploPoly(i).bSize,ExploPoly(i).bSize,1
        tbgl_CallList %LIST_EFFECT_QUAD
        billboard_End    
      tbgl_PopMatrix

    next  
    
    tbgl_Uselighting 1
    tbgl_UseBlend 0
    tbgl_UseDepthmask 1    
 
  end sub
  
  ' --
  
  sub Billboard_New( x as single, y as single, z as single, bSize as single, bSizeMax as single = -1, lifeSpan as long = -1, texture as long )
    BillboardID = iif(BillboardID = %MAX_BILLBOARDS, 1, BillboardID+1)
    Billboard(BillboardID).x = x
    Billboard(BillboardID).y = y
    Billboard(BillboardID).z = z
    Billboard(BillboardID).bSize = bSize/2
    Billboard(BillboardID).t = texture
    Billboard(BillboardID).render = 1
    if lifeSpan = -1 then
      Billboard(BillboardID).startTime = -1
      Billboard(BillboardID).endTime = -1
    else
      Billboard(BillboardID).startTime = gettickcount      
      Billboard(BillboardID).endTime = Gettickcount + lifeSpan      
    end if    
    Billboard(BillboardID).bSizeMax = bSizeMax
    Billboard(BillboardID).bSizeMin = bSize
  end sub
  
  
  sub Billboard_RenderAll()
    local i as long
    local Fade, Time as long = gettickcount
    tbgl_UseDepthmask 0    
    tbgl_UseBlend 1
    tbgl_Uselighting 0

    for i = 1 to %MAX_BILLBOARDS
      if Billboard(i).render = 0 then iterate for
      if gettickcount > Billboard(i).endTime then Billboard(i).render = 0
      
      tbgl_Bindtexture Billboard(i).t
      if Billboard(i).endTime > 0 then
        Fade = 128-cos((Time-Billboard(i).StartTime)/(Billboard(i).endTime-Billboard(i).StartTime)*6.28)*127
        '256-((Time-Billboard(i).StartTime) * 255/ (Billboard(i).endTime-Billboard(i).StartTime) )
        tbgl_color Fade, Fade, Fade
        Billboard(i).bSize = Billboard(i).bSizeMin+(Time-Billboard(i).StartTime) * Billboard(i).bSizeMax/(Billboard(i).endTime-Billboard(i).StartTime)
      else  
        tbgl_color 255,255,255
      endif  
      
      tbgl_Pushmatrix
        tbgl_Translate Billboard(i).x, Billboard(i).y, Billboard(i).z
        billboard_Begin        
        tbgl_Scale Billboard(i).bSize,Billboard(i).bSize,1
        tbgl_CallList %LIST_EFFECT_QUAD
        billboard_End          
      tbgl_Popmatrix
    next  

    tbgl_Uselighting 1
    tbgl_UseBlend 0
    tbgl_UseDepthmask 1    

  end sub  
  ' --
  
  sub billboard_Begin() 
    
    local modelview(16) as single
    
    ' save the current modelview matrix
    tbgl_PushMatrix

    ' get the current modelview matrix
    glGetFloatv(%GL_MODELVIEW_MATRIX , modelview )

    ' undo all rotations
    ' beware all scaling is lost as well 
    ' It will set first 3x3 submatrix to:
    ' 1 0 0
    ' 0 1 0
    ' 0 0 1
    modelview(1) = 1
    modelview(2) = 0
    modelview(3) = 0
    modelview(5) = 0
    modelview(6) = 1
    modelview(7) = 0
    modelview(9) = 0
    modelview(10) = 0
    modelview(11) = 1

    ' set the modelview with no rotations
    glLoadMatrixf(modelview)

  end sub

  sub billboard_End()

    ' restore the previously 
    ' stored modelview matrix
    tbgl_PopMatrix
    
  end sub  